from functools import reduce
from typing import Dict, Optional
import pandas as pd
import sqlite3
import os
from pathlib import Path

from .constants import *
from . import orm

from . import utils
from typing import Any, Dict, List, Optional, Tuple


def conn_and_cursor():
    conn = sqlite3.connect(os.path.join(HOMEDIR, Path(PROJID).with_suffix(".db")))
    cursor = conn.cursor()
    return conn, cursor


def insert_context(cursor, context):
    # Recursive
    if isinstance(context, dict):
        parent_context_id = (
            insert_context(cursor, context["p_ctx"])
            if context["p_ctx"] is not None
            else None
        )
    else:
        parent_context_id = (
            insert_context(cursor, context.p_ctx) if context.p_ctx is not None else None
        )
    if isinstance(context, orm.Loop):
        cursor.execute(
            """INSERT INTO loops (ctx_id, p_ctx_id, l_name, iteration, l_value) VALUES (?, ?, ?, ?, ?)""",
            (
                int(context.ctx_id),
                parent_context_id,
                str(context.name),
                int(context.iteration) if context.iteration is not None else None,
                str(context.value) if context.value is not None else None,
            ),
        )
        return int(context.ctx_id)
    else:
        cursor.execute(
            """INSERT INTO loops (ctx_id, p_ctx_id, l_name, iteration, l_value) VALUES (?, ?, ?, ?, ?)""",
            (
                int(context["ctx_id"]),
                parent_context_id,
                str(context["name"]),
                int(context["iteration"]) if context["iteration"] is not None else None,
                str(context["value"]) if context["value"] is not None else None,
            ),
        )
        return int(context["ctx_id"])


def unpack(output_buffer, cursor):
    if not output_buffer:
        return
    for each in output_buffer:
        if isinstance(each, orm.Log):
            ctx_id = insert_context(cursor, each.ctx) if each.ctx is not None else None
            cursor.execute(
                """INSERT INTO logs (projid, tstamp, filename, ctx_id, value_name, value, value_type) VALUES (?, ?, ?, ?, ?, ?, ?)""",
                (
                    each.projid,
                    each.tstamp,
                    each.filename,
                    ctx_id,
                    each.name,
                    str(each.value),
                    each.type,
                ),
            )
        else:
            # Parse JSON dict
            ctx_id = (
                insert_context(cursor, each["ctx"]) if each["ctx"] is not None else None
            )
            cursor.execute(
                """INSERT INTO logs (projid, tstamp, filename, ctx_id, value_name, value, value_type) VALUES (?, ?, ?, ?, ?, ?, ?)""",
                (
                    each["projid"],
                    each["tstamp"],
                    each["filename"],
                    ctx_id,
                    each["name"],
                    str(each["value"]),
                    each["type"],
                ),
            )


def create_tables(cursor):
    cursor.execute(
        """
        CREATE TABLE IF NOT EXISTS loops (
            ctx_id INTEGER,
            p_ctx_id INTEGER,
            l_name TEXT,
            iteration INTEGER,
            l_value TEXT
        )
        """
    )

    cursor.execute(
        """
        CREATE TABLE IF NOT EXISTS logs (
            projid TEXT,
            tstamp TEXT,
            filename TEXT,
            ctx_id INTEGER,
            value_name TEXT,
            value TEXT,
            value_type INTEGER
        )
        """
    )


def deduplicate_table(cursor, table_name):
    # Create a temporary table to store unique rows
    cursor.execute(
        f"""CREATE TEMPORARY TABLE temp_table AS SELECT DISTINCT * FROM {table_name}"""
    )

    # Delete the original table's contents
    cursor.execute(f"""DELETE FROM {table_name}""")

    # Insert the unique rows back into the original table
    cursor.execute(f"""INSERT INTO {table_name} SELECT * FROM temp_table""")

    cursor.execute("DROP TABLE temp_table")


def read_from_logs(cursor, where_clause=None):
    if where_clause is None:
        cursor.execute("SELECT DISTINCT * FROM logs")
    else:
        cursor.execute(f"SELECT DISTINCT * FROM logs WHERE {where_clause}")
    return cursor.fetchall()


def read_known_tstamps(cursor):
    cursor.execute("SELECT DISTINCT tstamp FROM logs")
    return cursor.fetchall()


def query(cursor, user_query, aspandas=False):
    cursor.execute(user_query)
    res = cursor.fetchall()
    if res and aspandas:
        return utils.cast_dtypes(pd.DataFrame(res, columns=get_column_names(cursor)))
    elif res:
        return res


def get_column_names(cursor):
    column_names = [description[0] for description in cursor.description]
    return column_names


def pivot(conn, *args):
    def _pivot_star():
        df = pd.read_sql(
            "SELECT DISTINCT value_name FROM logs WHERE value_type = 1 AND ctx_id IS NULL",
            conn,
        )
        value_names = df["value_name"].values
        if len(value_names) == 0:
            print("No default values to pivot on")
            return pd.DataFrame()

        # Build the dynamic part of the SQL query
        dynamic_sql = ", ".join(
            [
                f"MAX(CASE WHEN value_name = '{value_name}' THEN value ELSE NULL END) AS '{value_name}'"
                for value_name in value_names
            ]
        )

        # Construct the final SQL query
        final_sql = f"""
        SELECT projid,
            tstamp,
            filename,
            {dynamic_sql}
        FROM logs
        WHERE value_type = 1 AND ctx_id IS NULL
        GROUP BY projid, tstamp, filename;
        """

        # Execute the final SQL query
        return pd.read_sql(
            final_sql,
            conn,
            parse_dates=[
                "tstamp",
            ],
            coerce_float=True,
        )

    if not args:
        return _pivot_star()

    dataframes = []
    loops = pd.read_sql("SELECT * FROM loops", conn, coerce_float=False)
    for value_name in args:
        logs = pd.read_sql(
            f'SELECT * FROM logs WHERE value_name = "{value_name}"',
            conn,
            parse_dates=["tstamp"],
            coerce_float=True,
        )
        logs = logs[["projid", "tstamp", "filename", "ctx_id", "value"]]
        logs = logs.rename(columns={"value": value_name})
        while logs["ctx_id"].notna().any():
            logs = pd.merge(
                left=loops,
                right=logs,
                how="inner",
                on=["ctx_id"],
            )
            loop_name = logs["l_name"].unique()[0]
            logs = logs.drop(
                columns=[
                    "l_name",
                ]
            )
            if logs["iteration"].notna().all():
                logs["iteration"] = logs["iteration"].astype(int)
            logs = logs.rename(
                columns={"iteration": loop_name, "l_value": f"{loop_name}_value"}
            )
            if logs[loop_name].isna().any():
                logs = logs.drop(columns=[loop_name])
            if logs[f"{loop_name}_value"].isna().all():
                logs = logs.drop(columns=[f"{loop_name}_value"])

            assert loop_name in logs or f"{loop_name}_value" in logs

            logs["ctx_id"] = logs["p_ctx_id"]
            logs = logs.drop(columns=["p_ctx_id"])

        logs = logs.drop(columns=["ctx_id"])
        dataframes.append(logs)

    def join_on_common_columns(df1, df2):
        common_columns = set(df1.columns) & set(df2.columns)
        return pd.merge(df1, df2, on=list(common_columns), how="outer")

    all_joined = reduce(join_on_common_columns, dataframes)
    cols = [c for c in all_joined.columns if c not in args]
    cols += list(args)
    return all_joined[cols]



# type: ignore

from ast import AST, iter_fields
from typing import Any, Iterable, Optional, Union

from .adapter import BaseAdapter, memoize, materialize

# https://github.com/PyCQA/pylint/issues/3882
# pylint: disable=unsubscriptable-object

from ast import expr_context, boolop, operator, unaryop, cmpop

Enums = (expr_context, boolop, operator, unaryop, cmpop)


Node = Union[AST, list]


class Adapter(BaseAdapter):
    def __init__(self, *roots: Node):
        """
        DFS tree generator
        implements:
          self[Tree].children(Node)
          self[Tree].label(Node)
          self[Tree[.value(Node)
        """

        self._parents = {}
        for root in roots:
            self._save_parents(root)

    def parent(self, n: Node) -> Optional[Node]:
        return self._parents[id(n)]

    def _save_parents(self, node, parent=None):
        assert id(node) not in self._parents
        self._parents[id(node)] = parent
        for child in self.children(node):
            self._save_parents(child, parent=node)

    @memoize
    @materialize
    def children(self, n: Node) -> Iterable[Node]:
        if isinstance(n, AST):
            it = iter_fields(n)
        if isinstance(n, list):
            it = enumerate(n)

        for _, value in it:
            if isinstance(value, (AST, list)) and not isinstance(value, Enums):
                yield value

    @memoize
    def label(self, n: Node) -> str:
        return type(n).__name__

    @memoize
    def value(self, n: Node) -> Any:
        terminals = []
        if isinstance(n, AST):
            for name, value in iter_fields(n):
                if (
                    isinstance(value, list)
                    and value
                    and all(isinstance(e, Enums) for e in value)
                ):
                    terminals.append((name, value))
                if (
                    value is not None
                    and not isinstance(value, (AST, list))
                    or isinstance(value, Enums)
                ):
                    terminals.append((name, value))
        return terminals



import json
from typing import Any
from pathlib import Path

import pandas as pd

import warnings

warnings.filterwarnings(
    "ignore", category=UserWarning, message="Could not infer format.*"
)


def is_jsonable(x):
    try:
        json.dumps(x)
        return True
    except (TypeError, OverflowError):
        return False


def duck_cast(v: str, default: Any):
    if isinstance(default, bool):
        return v.lower() in ("yes", "y", "true", "t", "1")
    if isinstance(default, int):
        return int(v)
    elif isinstance(default, float):
        return float(v)
    elif isinstance(default, str):
        return str(v) if v else ""
    elif isinstance(default, list):
        return list(v) if v else []
    elif isinstance(default, tuple):
        return tuple(v) if v else tuple([])
    else:
        raise TypeError(f"Unsupported type: {type(default)}")


def add2copy(src, name, value):
    d = dict(src)
    d[name] = value
    return d


def to_string(src, name, value):
    if src:
        return f"{', '.join([f'{k}: {v}' if v is not None else f'{k}: {i}' for k,(i,v) in src.items()])}, {name}: {str(value)}"
    else:
        return f"{name}: {str(value)}"


def to_filename(src, name, ext):
    rolling_name = [
        name,
    ]
    for k, (i, v) in src.items():
        rolling_name.append(str(k))
        if v is not None:
            rolling_name.append(str(v))
        else:
            rolling_name.append(str(i))
    return Path("_".join(rolling_name)).with_suffix(ext)


def split_and_retrieve_elements(array, count=10):
    # Find the middle index
    middle_index = len(array) // 2

    # Split the array into two halves
    left_half = array[:middle_index]
    right_half = array[middle_index:]

    # Retrieve the first 10 elements from the left half and last 10 elements from the right half
    first_10_left = left_half[:count]
    last_10_right = right_half[-count:]

    return first_10_left, last_10_right


def is_integer(string):
    try:
        int(string)
        return True
    except ValueError:
        return False


def cast_dtypes(df: pd.DataFrame, columns=None):
    target_columns = columns if columns is not None else df.columns
    for col in target_columns:
        if df[col].dtype == "object":
            # Attempt to convert to numeric first
            try:
                df[col] = pd.to_numeric(df[col], errors="raise")
            except ValueError:
                # If it fails, try coerce (turns non-numeric into NaN)
                pass

            # After numeric attempt, if still object, try datetime
            if df[col].dtype == "object":
                try:
                    df[col] = pd.to_datetime(df[col], errors="raise")
                except ValueError:
                    # If datetime also fails, revert to no conversion or coerce
                    pass
    return df


def latest(df: pd.DataFrame):
    # return df where tstamp is the latest tstamp (and only the latest tstamp)
    if not df.empty:
        return df[df["tstamp"] == df["tstamp"].max()]
    return df


def discretize(cost_estimate: float):
    if cost_estimate < 10:
        return "under 10 seconds"
    elif cost_estimate < 100:
        return "under 2 minutes"
    elif cost_estimate < 1000:
        return "up to 15 minutes"
    else:
        return "more than 15 minutes"



import json
import argparse
from argparse import Namespace
from typing import Any, Dict, Optional, Tuple

from dataclasses import dataclass
from .versions import current_branch, to_shadow
import sys

from .hlast.visitors import WithExpVisitor
import ast


@dataclass
class Flags:
    hyperparameters: Dict[str, str]
    queryparameters: Optional[Dict[str, str]]
    old_tstamp: Optional[str]
    args: Optional[Any]
    columns: Optional[Tuple[str]]


flags = Flags({}, None, None, None, None)


def parse_replay_flor(arg):
    parts = arg.split()
    d = {}
    for part in parts:
        if "=" not in part:
            d["VARS"] = part.split(",")
        else:
            key, value = part.split("=")
            d[key] = value.split(",")
    return d


def parse_columns(column_string):
    return [str(each) for each in column_string.split()]


def parse_args():
    parser = argparse.ArgumentParser(description="FlorDB CLI")

    # Existing code
    parser.add_argument(
        "--replay_flor",
        nargs="*",
        type=parse_replay_flor,
        help="Key-value pair arguments corresponding to `flor.loop` name and access method",
    )
    parser.add_argument(
        "--kwargs",
        nargs="*",
        type=str,
        help="Additional key-value pair arguments for hyper-parameters",
    )

    # Flor module subparser
    flor_parser = parser.add_subparsers(dest="flor_command")

    # Unpack command
    unpack_parser = flor_parser.add_parser("unpack")
    stat_parser = flor_parser.add_parser("stat")

    # Replay command
    replay_parser = flor_parser.add_parser(
        "replay", help="Replay with specified VARS and optional QUERY"
    )
    replay_parser.add_argument(
        "VARS",
        type=lambda s: s.split(","),
        help="List of logged variables, or linenos, comma-separated",
    )
    replay_parser.add_argument(
        "where_clause", nargs="?", type=str, help="Optional SQL WHERE clause"
    )

    # Query command
    query_parser = flor_parser.add_parser("query")
    query_parser.add_argument(
        "q", type=str, help="SQL query to execute on the database"
    )

    # Pivot command
    pivot_parser = flor_parser.add_parser("dataframe")
    pivot_parser.add_argument(
        "columns",
        nargs="?",
        type=lambda s: s.split(","),
        help="List of logged variables, comma-separated",
    )

    flor_commands = [
        "--kwargs",
        "--replay_flor",
        "unpack",
        "replay",
        "query",
        "dataframe",
        "stat",
    ]

    # Check if any of the flor_commands is in the arguments
    if any(command in sys.argv for command in flor_commands):
        args = parser.parse_args()
        flags.args = args
    else:
        flags.args = None

    if flags.args is not None and flags.args.replay_flor is not None:
        flags.queryparameters = {}
        for d in flags.args.replay_flor:
            flags.queryparameters.update(d)
        replay_initialize()

    # Process the key-value pair arguments
    if flags.args is not None and flags.args.kwargs is not None:
        if not flags.args.kwargs:
            raise RuntimeError("--kwargs called but no arguments added")
        for kwarg in flags.args.kwargs:
            key, value = kwarg.split("=")
            flags.hyperparameters[key] = value

    if flags.args is not None and flags.args.flor_command == "dataframe":
        flags.columns = flags.args.columns

    return flags


def in_replay_mode():
    return flags.queryparameters is not None


def replay_initialize():
    assert (
        flags.args is not None and flags.args.kwargs is None
    ), "Cannot set --kwargs in replay, would rewrite history"
    # read .flor_replay.json into flags.args.VARS
    # update flags.hyperparameters
    with open(".flor.json", "r") as f:
        data = json.load(f)
        filename = data[0]["filename"]

    with open(filename, "r") as f:
        tree = ast.parse(f.read())
    wev = WithExpVisitor()
    wev.visit(tree)

    flags.queryparameters["WEV"] = wev.found

    for obj in data:
        if obj["ctx"] is None and obj["type"] == 1:
            d = {obj["name"]: obj["value"]}
            flags.hyperparameters.update(d)
    flags.old_tstamp = data[0]["tstamp"]



import ast
import json
import re
import shutil
import numpy as np
import pandas as pd
from typing import List, Optional
import subprocess
import tempfile
import os

from . import utils
from .hlast.visitors import LoggedExpVisitor, WithExpVisitor
from .hlast import backprop

from . import database
from . import versions
from .clock import Clock


def dataframe(*args):
    conn, _ = database.conn_and_cursor()
    # Query the distinct value_names
    try:
        df = database.pivot(conn, *(args if args else tuple()))
        df = df.drop_duplicates()
        df = df.reset_index(drop=True)
        return df
    finally:
        conn.close()


def query(user_query: str):
    conn, cursor = database.conn_and_cursor()
    try:
        df = database.query(cursor, user_query, aspandas=True)
        return df
    finally:
        # Close connection
        conn.commit()
        conn.close()


def replay(apply_vars: List[str], where_clause: Optional[str] = None):
    versions.git_commit("Hindsight logging stmts added.")
    schedule = Schedule(apply_vars, where_clause)

    with open(".flor.json", "r") as f:
        main_script = json.load(f)[0]["filename"]
    temp_file = tempfile.NamedTemporaryFile(delete=False)
    shutil.copy2(main_script, temp_file.name)
    with open(main_script, "r") as f:
        tree = ast.parse(f.read())
    lev, wev = LoggedExpVisitor(), WithExpVisitor()
    wev.visit(tree)
    lev.visit(tree)

    if not wev.found:
        # "No `with flor.checkpointing(...):` statement found in main script."
        loglvl, mark = 3, "suffix"
    else:
        loglvl, mark = schedule.get_loglvl(lev)

    assert mark in ("prefix", "suffix")
    level_mapper = {0: "prefix", 1: "outer loop", 2: "nested loop", 3: "full scan"}

    schedule.estimate_cost(loglvl, mark)

    if loglvl == 3:
        print("log level full scan, replaying from flor.args.")
    elif loglvl < 3:
        print(
            "log level",
            level_mapper[loglvl],
            "to suffix." if mark == "suffix" else "without suffix.",
        )
    print()
    print(schedule.df)
    print()

    res = input(
        f"Continue replay estimated to finish in {utils.discretize(sum(schedule.df['composite']))} [y/N]? "
    )
    res = res if res else "n"
    if res.lower().strip() == "n":
        return schedule

    clock = Clock()
    clock.set_start_time()

    # Pick up on versions
    active_branch = versions.current_branch()
    try:
        for projid, ts, hexsha, main_script in schedule.iter_dims():
            print("entering", str(ts), hexsha)
            versions.checkout(hexsha)
            for v, lineno in zip(
                apply_vars,
                [int(v) if utils.is_integer(v) else lev.names[v] for v in apply_vars],
            ):
                print("applying: ", v, lineno)
                try:
                    backprop(lineno, temp_file.name, main_script, main_script)
                except Exception as e:
                    print("Exception raised during `backprop`", e)
                    raise e
            if loglvl == 0 or loglvl == 3:
                print("loglvl", loglvl)
                cmd = ["python", main_script, "--replay_flor", ",".join(apply_vars)]
                print(*cmd)
                subprocess.run(cmd)
            elif loglvl == 1:
                tup = (
                    ",".join(
                        [
                            str(i)
                            for i in range(
                                schedule.df[schedule.df["tstamp"] == ts][
                                    "num_epochs"
                                ].values[0]
                            )
                        ]
                    )
                    + ","
                )
                print("loglvl", loglvl, tup)
                cmd = ["python", main_script, "--replay_flor", ",".join(apply_vars)] + [
                    "epoch=" + tup
                ]
                print(*cmd)
                subprocess.run(cmd)
            elif loglvl == 2:
                tup = (
                    ",".join(
                        [
                            str(i)
                            for i in range(
                                schedule.df[schedule.df["tstamp"] == ts][
                                    "num_epochs"
                                ].values[0]
                            )
                        ]
                    )
                    + ","
                )
                print("loglvl", loglvl, tup)
                cmd = ["python", main_script, "--replay_flor", ",".join(apply_vars)] + [
                    "epoch=" + tup,
                    "step=1",
                ]
                print(*cmd)
                subprocess.run(cmd)
            else:
                raise NotImplementedError(
                    "Please open a Pull Request on GitHub and describe your use-case."
                )
    except Exception as e:
        print("Exception raised during `schedule.iter_dims()`", e)
        raise e
    finally:
        versions.reset_hard()
        versions.checkout(active_branch)
        os.remove(temp_file.name)

    dt = clock.get_delta()

    filtered_vs = [v for v in apply_vars if not utils.is_integer(v)]
    if schedule.vars_in_where is not None:
        filtered_vs += schedule.vars_in_where
    schedule = dataframe(*filtered_vs)

    print()
    print(schedule)
    print()
    print(dt, "seconds")

    return schedule


class Schedule:
    def __init__(self, apply_vars, where_clause) -> None:
        # TODO:
        # case when integer supplied through apply_vars,
        #     you will need to infer var_name from ast
        self.apply_vars = apply_vars
        self.where_clause = where_clause
        self.vars_in_where = None
        if where_clause is not None:
            # Regular expression to match column names
            column_pattern = re.compile(r"\b[A-Za-z_]\w*\b")
            columns = set(re.findall(column_pattern, where_clause))

            # Convert to list if needed
            columns_list = list(columns)
            self.vars_in_where = columns_list
            print("columns in where_clause:", columns_list)

    def estimate_cost(self, loglvl: int, mark: str):
        assert mark in ("prefix", "suffix")
        keys = ["projid", "tstamp", "filename"]
        if loglvl == 3:
            # Full scan, delta::suffix is the estimate per run
            pvt = dataframe()
            self.df = dataframe("delta::suffix")
            self.df["composite"] = pd.to_numeric(self.df["delta::suffix"])
            self.df = pd.merge(pvt, self.df, on=keys, how="inner")
        else:
            pvt = dataframe()
            if loglvl == 0:
                if mark == "prefix":
                    self.df = dataframe("delta::prefix")
                    self.df["composite"] = pd.to_numeric(self.df["delta::prefix"])
                else:
                    self.df = dataframe("delta::prefix", "delta::suffix")
                    self.df["composite"] = (
                        pd.to_numeric(self.df["delta::prefix"])
                        + pd.to_numeric(self.df["delta::suffix"])
                        + 1
                    )
                self.df = pd.merge(pvt, self.df, on=keys, how="inner")
            elif loglvl == 1:
                # i - delta::loop where i is the full duration for that iteration
                df = dataframe("delta::loop")
                df["delta::loop"] = pd.to_numeric(df["delta::loop"], errors="coerce")
                df_grouped = (
                    df.groupby(keys)
                    .agg(
                        num_epochs=("epoch", "max"),
                        sum_nested_loops=("delta::loop", "sum"),
                    )
                    .reset_index()
                )
                temp_df = query(
                    "SELECT * FROM logs WHERE ctx_id is null and value_name='delta::loop';"
                )
                temp_df.drop(columns=["ctx_id", "value_name", "value_type"], inplace=True)  # type: ignore
                temp_df = temp_df.rename(columns={"value": "coarse_loop"})  # type: ignore
                temp_df["coarse_loop"] = pd.to_numeric(
                    temp_df["coarse_loop"], errors="coerce"
                )

                merged_df = pd.merge(temp_df, df_grouped, on=keys, how="inner")
                merged_df["marginal"] = (
                    merged_df["coarse_loop"] - merged_df["sum_nested_loops"]
                )
                merged_df.drop(
                    columns=["coarse_loop", "sum_nested_loops"], inplace=True
                )
                df = dataframe("delta::prefix", "delta::suffix")
                merged_df = pd.merge(merged_df, df, on=keys, how="inner")
                merged_df["composite"] = (
                    pd.to_numeric(merged_df["num_epochs"])
                    + merged_df["marginal"]
                    + pd.to_numeric(merged_df["delta::prefix"])
                    + pd.to_numeric(merged_df["delta::suffix"])
                )
                self.df = pd.merge(pvt, merged_df, on=keys, how="inner")
            elif loglvl == 2:
                base_df = dataframe("delta::prefix", "delta::suffix")
                loop_df = dataframe("delta::loop")
                loop_df["delta::loop"] = pd.to_numeric(
                    loop_df["delta::loop"], errors="coerce"
                )
                df_grouped = (
                    loop_df.groupby(keys).agg(num_epochs=("epoch", "max")).reset_index()
                )
                temp_df = query(
                    "SELECT * FROM logs WHERE ctx_id is null and value_name='delta::loop';"
                )
                temp_df.drop(columns=["ctx_id", "value_name", "value_type"], inplace=True)  # type: ignore
                temp_df = temp_df.rename(columns={"value": "coarse_loop"})  # type: ignore
                temp_df["coarse_loop"] = pd.to_numeric(
                    temp_df["coarse_loop"], errors="coerce"
                )
                merged_df = pd.merge(temp_df, base_df, on=keys, how="inner")
                merged_df["composite"] = (
                    pd.to_numeric(merged_df["coarse_loop"])
                    + pd.to_numeric(merged_df["delta::prefix"])
                    + pd.to_numeric(merged_df["delta::suffix"])
                )
                merged_df = pd.merge(pvt, merged_df, on=keys, how="inner")
                merged_df = pd.merge(merged_df, df_grouped, on=keys, how="inner")
                self.df = merged_df
            else:
                raise

    def is_empty(self):
        return self.df.empty

    def get_loglvl(self, lev: LoggedExpVisitor):
        # Get the largest lineno from self.apply_vars
        max_lineno = max(lev.names[v] for v in self.apply_vars)
        loglevels = [lev.line2level[lev.names[v]] for v in self.apply_vars]

        # Check for monotonic growth
        pairs = sorted(
            [
                (line, level)
                for line, level in lev.line2level.items()
                if line <= max_lineno
            ],
            key=lambda x: x[0],
        )
        is_monotonic = all(
            pairs[i][1] <= pairs[i + 1][1] for i in range(len(pairs) - 1)
        )

        return (
            max(loglevels),
            "prefix" if is_monotonic else "suffix",
        )

    def iter_dims(self):
        ts2vid = {
            pd.Timestamp(ts): str(vid)
            for ts, vid, _ in versions.get_latest_autocommit()
        }

        prev_row = None

        for row_dict in self.df.to_dict(orient="records"):
            curr_tstamp = row_dict["tstamp"]

            # Compare current timestamp with the previous timestamp
            if prev_row is not None and curr_tstamp != prev_row["tstamp"]:
                yield prev_row["projid"], prev_row["tstamp"], ts2vid[
                    prev_row["tstamp"]
                ], prev_row["filename"]

            # Update prev_row for the next iteration
            prev_row = row_dict

        # Yield the final record if prev_row is populated
        if prev_row is not None:
            yield prev_row["projid"], prev_row["tstamp"], ts2vid[
                prev_row["tstamp"]
            ], prev_row["filename"]

    def __str__(self):
        if self.where_clause is None:
            return self.df[self.df[self.apply_vars].isna().any(axis=1)].__str__()  # type: ignore
        else:
            schedule = self.df.copy()  # appease pandas warning
            schedule[[v for v in self.apply_vars if v not in schedule.columns]] = np.nan
            return schedule.__str__()

    def __repr__(self):
        if self.where_clause is None:
            return self.df[self.df[self.apply_vars].isna().any(axis=1)].__repr__()  # type: ignore
        else:
            schedule = self.df.copy()  # appease pandas warning
            schedule[[v for v in self.apply_vars if v not in schedule.columns]] = np.nan
            return schedule.__repr__()

    def _repr_html_(self):
        if self.where_clause is None:
            return self.df[self.df[self.apply_vars].isna().any(axis=1)]._repr_html_()  # type: ignore
        else:
            schedule = self.df.copy()  # appease pandas warning
            schedule[[v for v in self.apply_vars if v not in schedule.columns]] = np.nan
            return schedule._repr_html_()  # type: ignore



from dataclasses import dataclass, asdict
from typing import Any, List, Optional
import json
import os
import random

from . import versions


def generate_64bit_id() -> int:
    return random.randint(-(2**63), 2**63 - 1)


@dataclass
class Loop:
    ctx_id: int
    p_ctx: Optional["Loop"]
    name: str
    iteration: int
    value: str


@dataclass
class Log:
    projid: str
    tstamp: str
    filename: str
    ctx: Optional[Loop]
    name: str
    value: Any
    type: int


def to_json(output_buffer: List[Log]):
    buffer = [asdict(o) for o in output_buffer]
    with open(os.path.join(versions.get_repo_dir(), ".flor.json"), "w") as f:
        json.dump(buffer, f, indent=2)



from _ast import AST, Constant
from typing import Any, Dict, Optional
import ast

from .. import utils


class WithExpVisitor(ast.NodeVisitor):
    def __init__(self):
        super().__init__()
        self.found = False

    def visit_With(self, node: ast.With):
        """
        Checks for the existence of a with statement:
        with flor.checkpointing(...):
            ...
        """
        pred = (
            isinstance(node.items[0].context_expr, ast.Call)
            and isinstance(node.items[0].context_expr.func, ast.Attribute)
            and isinstance(node.items[0].context_expr.func.value, ast.Name)
            and node.items[0].context_expr.func.value.id == "flor"
            and node.items[0].context_expr.func.attr == "checkpointing"
        )
        if pred:
            self.found = True
        else:
            self.generic_visit(node)


class LoggedExpVisitor(ast.NodeVisitor):
    def __init__(self):
        super().__init__()
        self.names: Dict[str, int] = {}

        self.line2level: Dict[int, int] = {}
        self.lvl = 0

    def visit_For(self, node: ast.For):
        iter_s = ast.unparse(node.iter).strip()
        if iter_s.startswith("flor.loop"):
            start_lvl = self.lvl
            try:
                self.lvl += 1
                self.generic_visit(node)
            finally:
                self.lvl = start_lvl
        else:
            self.generic_visit(node)

    def visit_Call(self, node: ast.Call):
        pred = (
            isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Name)
            and node.func.value.id == "flor"
            and node.func.attr == "log"
        )
        if not pred:
            return self.generic_visit(node)
        if len(node.args) == 2 and isinstance(node.args[0], ast.Constant):
            self.names[str(node.args[0].value)] = node.lineno
            self.line2level[node.lineno] = self.lvl
        else:
            raise IndexError("FLOR: Did you give flor.log a key? It takes 2 args.")

    def generic_visit(self, node: AST) -> Any:
        if hasattr(node, "lineno"):
            self.line2level[node.lineno] = self.lvl
        return super().generic_visit(node)


class NoGradVisitor(ast.NodeVisitor):
    def __init__(self):
        super().__init__()
        self.feeding = False
        self.names = {}
        self.tree = None

    def visit_Call(self, node: ast.Call):
        pred = (
            isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Name)
            and node.func.value.id == "flor"
            and node.func.attr == "log"
        )
        if not pred or not self.feeding:
            return self.generic_visit(node)
        if len(node.args) == 2 and isinstance(node.args[0], ast.Constant):
            self.names[str(node.args[0].value)] = node.lineno
        else:
            raise IndexError("FLOR: Did you give flor.log a key? It takes 2 args.")

    def visit_With(self, node: ast.With):
        if [True for each in node.items if "torch.no_grad" in ast.unparse(each)]:
            try:
                feeding = self.feeding
                self.feeding = True
                self.tree = node
                for stmt in node.body:
                    self.visit(stmt)
            finally:
                self.feeding = feeding  # type: ignore


class NoGradTransformer(ast.NodeTransformer):
    def __init__(self, old_tree) -> None:
        super().__init__()
        self.their_tree = old_tree

    def visit_With(self, node: ast.With):
        if [True for each in node.items if "torch.no_grad" in ast.unparse(each)]:
            return self.generic_visit(self.their_tree)
        else:
            return self.generic_visit(node)



from argparse import Namespace

from .gtpropagate import propagate


def backprop(lineno: int, source, target, out=None):
    try:
        if out is None:
            with open(str(source), "r") as src, open(str(target), "r") as dst:
                return propagate(
                    Namespace(
                        lineno=lineno,
                        source=src,
                        target=dst,
                        out=str(target),
                        gumtree=dict(),
                    )
                )
        else:
            with open(str(source), "r") as src, open(str(target), "r") as dst:
                return propagate(
                    Namespace(
                        lineno=lineno, source=src, target=dst, out=out, gumtree=dict()
                    )
                )
    except FileExistsError:
        print("Logging record exists in target, nothing to do")


__all__ = ["backprop"]

if __name__ == "__main__":
    backprop(78, "cases/train_rnn/now.py", "cases/train_rnn/before.py")



import setuptools  # type: ignore
import io

with io.open("README.md", mode="r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="flordb",
    version="3.4.12",
    author="Rolando Garcia",
    author_email="rolando.garcia@asu.edu",
    description="A hindsight logging dataabase for MLOps",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ucbrise/flor",
    packages=setuptools.find_packages(),
    install_requires=[
        "GitPython",
        "cloudpickle",
        "astunparse",
        "pandas",
        "bidict==0.21.3",
        "apted",
        "matplotlib",
        "scikit-learn",
        "numpy",
        "tqdm",
        "sh",
        "ipython",
        "ipykernel",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9",
)



from apted import APTED, Config
from itertools import product
from collections.abc import Iterable
from typing import Generic, Optional, TypeVar
from bidict._mut import MutableBidict

from .adapter import Adapter
from .idmap import IdMap
from .priorityq import PriorityQ

# https://github.com/PyCQA/pylint/issues/3882
# pylint: disable=unsubscriptable-object


Tree = TypeVar("Tree")


class HeightPQ(PriorityQ[Tree, int]):
    def __init__(self, adapter: Adapter[Tree], it=[]):
        super().__init__(it, key=adapter.height, reverse=True)  # type: ignore
        self.adapter = adapter

    def peek_max(self) -> int:
        return len(self) and self.adapter.height(self.peek())

    def open(self, tree: Tree):
        for child in self.adapter.children(tree):
            self.push(child)

    def pop(self) -> list[Tree]:
        trees = []
        assert self, "Empty!"
        height = self.peek_max()
        while self.peek_max() == height:
            trees.append(super().pop())
        return trees


class Mapping(MutableBidict[Tree, Tree]):  # type: ignore
    _fwdm_cls = IdMap[Tree, Tree]  # type: ignore
    _invm_cls = IdMap[Tree, Tree]  # type: ignore
    _repr_delegate = list

    def __init__(self, adapter: Adapter[Tree], it: Iterable[tuple[Tree, Tree]] = ()):
        self.adapter = adapter
        super().__init__(it)

    def put_tree(self, t1: Tree, t2: Tree):
        self.putall(zip(*map(self.adapter.postorder, [t1, t2])))  # type: ignore


class GumTree(Generic[Tree]):

    """

    Fine-grained and accurate source code differencing
    Falleri et al., ASE 2014.

    """

    defaults = {"min_height": 2, "min_dice": 0.50, "max_size": 100}

    def __init__(self, adapter: Adapter[Tree], *, opt=None, **params):
        assert not set(params) - set(self.defaults), "Invalid parameters!"
        self.params = dict(self.defaults, **params)
        self.opt = opt or self.apted
        self.adapter = adapter

    def mapping(self, t1: Tree, t2: Tree) -> Mapping[Tree, Tree]:  # type: ignore
        m = self.topdown(t1, t2)
        self.bottomup(t1, t2, m)
        return m

    def topdown(self, t1: Tree, t2: Tree) -> Mapping[Tree, Tree]:  # type: ignore
        min_height = self.params["min_height"]

        parent = self.adapter.parent
        isomorphic = self.adapter.isomorphic

        def different(l, r):
            return id(l) != id(r)

        adapter = self.adapter
        l1, l2 = HeightPQ(adapter, [t1]), HeightPQ(adapter, [t2])
        a, m = [], Mapping(adapter)

        # Note: Algorithm uses >, but the example seems to use >= instead.
        while max(l.peek_max() for l in [l1, l2]) >= min_height:
            if l1.peek_max() != l2.peek_max():
                pq = max(l1, l2, key=HeightPQ.peek_max)
                for t in pq.pop():
                    pq.open(t)
            else:
                h1, h2 = l1.pop(), l2.pop()
                for n1, n2 in product(h1, h2):
                    if isomorphic(n1, n2):
                        if any(
                            isomorphic(n1, t) for t in h2 if different(t, n2)
                        ) or any(isomorphic(t, n2) for t in h1 if different(t, n1)):
                            a.append((n1, n2))
                        else:
                            m.put_tree(n1, n2)
                for t1 in h1:
                    if t1 not in m:
                        l1.open(t1)
                for t2 in h2:
                    if t2 not in m.inv:
                        l2.open(t2)

        a.sort(key=lambda ts: self.dice(*map(parent, ts), m), reverse=True)  # type: ignore
        for n1, n2 in a:
            # Note: Algorithm removes from A, but I think this is more efficient
            if n1 not in m and n2 not in m.inv:
                m.put_tree(n1, n2)

        return m

    def bottomup(self, t1: Tree, t2: Tree, m: Mapping[Tree, Tree]):  # type: ignore
        min_dice = self.params["min_dice"]
        max_size = self.params["max_size"]

        label = self.adapter.label
        postorder = self.adapter.postorder
        num_descendants = self.adapter.num_descendants

        # FIXME: Paper mentions candidates must have descendants matched, but
        # it was hard to do and if min_dice > 0 they will be dropped anyways.
        assert min_dice > 0

        def candidate(n1: Tree, m: Mapping[Tree]):  # type: ignore
            return max(
                (
                    c2
                    for c2 in postorder(t2)
                    if label(n1) == label(c2) and c2 not in m.inv
                ),
                key=lambda c2: self.dice(n1, c2, m),
                default=None,
            )

        for n1 in postorder(t1):
            if n1 not in m:
                n2 = candidate(n1, m)
                if n2 and self.dice(n1, n2, m) > min_dice:
                    m.put(n1, n2)
                    # Note: Paper mentions removing already matched descendants
                    if max(num_descendants(t) for t in [n1, n2]) < max_size:
                        for ta, tb in self.opt(n1, n2):
                            if (
                                ta is not None
                                and tb is not None
                                and ta not in m
                                and tb not in m.inv
                                and label(ta) == label(tb)
                            ):
                                m.put(ta, tb)

    def dice(self, t1: Tree, t2: Tree, m: Mapping[Tree]):  # type: ignore
        num_descendants = self.adapter.num_descendants
        descendants = self.adapter.descendants
        contains = self.adapter.contains

        # Note: Formula is unclear, I think this is what they meant ¯\_(ツ)_/¯
        return (
            2
            * sum(1 for n1 in descendants(t1) if n1 in m and contains(m[n1], t2))
            / (num_descendants(t1) + num_descendants(t2) or 1)
        )

    def apted(self, t1: Tree, t2: Tree):
        return APTED(t1, t2, AptedConfig(self.adapter)).compute_edit_mapping()


class AptedConfig(Config):
    def __init__(self, adapter):
        self.adapter = adapter

    def rename(self, n1: Tree, n2: Tree):
        return int(self.adapter.label(n1) != self.adapter.label(n2))

    def children(self, n):
        return list(self.adapter.children(n))
